
#include "dispatch.h"
/*
Setup mode / surveilannce combined cam
  
Adds new sources as new sources register with the image db.
(Good debug would be to write the filenames using OSD
*/

#define MUNIN_ENABLED 1
#define MUNIN_STORAGE "deja_images"

#define DB_SIZE  2048
#define DB_SLACK 256
#define SAVE_INTERVAL 512

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include "db.h"
#include "capture.h"
#include "load_image.h"
#include "digest.h"
#include <SDL.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include "lost-in-time.h"

#define USE_OVERLAY

static int SUBDIV         = 1;
static int WINDOW_WIDTH   = 1024;
static int WINDOW_HEIGHT  = 768;
static int USE_FULLSCREEN = 0;


static void parse_args (int    argc, char **argv);
static char *active_stream = NULL;

static int  event_iteration ();
static int  memory_iteration (void *db);
static akois_db * memory_init (void);
static void memory_destroy ();
static void deja_init ();
static void deja_destroy ();
static void video_init ();
static void video_destroy ();

static int running=1;

extern int CAPTURE_WIDTH;
extern int CAPTURE_HEIGHT;

int
munin_deja (int argc,
            char **argv)
{
  Dispatch *dispatch;
  akois_db *db;
  int cap_w = CAPTURE_WIDTH;
  int cap_h = CAPTURE_HEIGHT;

  CAPTURE_WIDTH = 640;
  CAPTURE_HEIGHT = 480;
  
  parse_args (argc, argv);
  if (!active_stream)
    active_stream = strdup ("cam0");


  dispatch = dispatch_new ();
  db=memory_init ();

  {
	  char buf[256];
	  sprintf (buf, "install -d %s", MUNIN_STORAGE);
	  system (buf);
  }
  //capture_set_mask (1);
  capture_set_dir ("/tmp/munin_deja/");
  dispatch_add_task (dispatch, 0, video_init, event_iteration,  video_destroy,  NULL);
  dispatch_add_task (dispatch, 0, capture_init, capture_iteration,  capture_destroy,  NULL);
  dispatch_add_task (dispatch, 0, NULL,       memory_iteration, memory_destroy, db);
  dispatch_add_task (dispatch, 0, deja_init,  NULL,             deja_destroy,   NULL);

  running=1;
  while (dispatch_iterate (dispatch) && running);

  dispatch_destroy (dispatch);

  CAPTURE_WIDTH = cap_w;
  CAPTURE_HEIGHT = cap_h;
  return 0;
}


#include <oxide/oxide.h>
Oxide *oxide = NULL;
Oxide *scene_detector;

#define hor_tiles SUBDIV
#define ver_tiles SUBDIV

#define SCREEN_FLAGS   (SDL_RESIZABLE | SDL_HWSURFACE | (USE_FULLSCREEN?SDL_FULLSCREEN:0))
#define SCREEN_BPP     32

#include <gggl/image/gggl_image.h>

GgglImage *temp_image = NULL;

static SDL_Surface *screen = NULL;
#ifdef USE_OVERLAY
static SDL_Overlay *bmp = NULL;
#endif

static akois_db *db;  /* to have it available for the SIGINT proc */

static void
munin_display_buf (const unsigned char *pixels,
                   int   width,
                   int   height,
                   int   row_stride,
                   int   tile_x,
                   int   tile_y);


static void save_db()
{
  system ("cp deja.db deja.db~");
  akois_db_save (db, "deja.db");
}

static void dump_db()
{
  fprintf (stderr, "\nMunin is terminating\n");
  SDL_Quit ();
  save_db ();
  akois_db_free (db);
  //exit (0);
}

static void
munin_display (const char *file,
               int         tile_x,
               int         tile_y)
{
  unsigned char *pixels;
  int width;
  int height;
  int row_stride;

  if (!load_image (file, &width, &height, &row_stride, &pixels))
    {
      munin_display_buf (pixels, width, height, row_stride,
                         tile_x, tile_y);
      free (pixels);
    }
}

static void
munin_query (akois_db      *db,
             akois_image   *image,
             akois_options *options)
{
  int x,y;

  for (y=0;y<ver_tiles;y++)
    {
#ifdef USE_OVERLAY
  SDL_LockYUVOverlay (bmp);
#endif
#ifndef USE_OVERLAY
  SDL_LockSurface (screen);
#endif
    for (x=0;x<hor_tiles;x++)
      {
         if (!running)
           return;
      
         akois_image *result;
         akois_image_calc_digest (image);
         result = akois_db_best (db, image, options);
         if (result)
            munin_display (akois_image_get_path (result), x, y);
      }
#ifdef USE_OVERLAY
  SDL_UnlockYUVOverlay (bmp);

  {
    SDL_Rect  rect;

    rect.x = 0;
    rect.y = 0;
    rect.w = WINDOW_WIDTH;
    rect.h = WINDOW_HEIGHT;

    SDL_DisplayYUVOverlay (bmp, &rect);
  }
#endif
#ifndef USE_OVERLAY
  SDL_UnlockSurface (screen);
  SDL_UpdateRect(screen, 0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
#endif
    }
  akois_db_reduce_to (db, DB_SIZE, DB_SLACK);
}

akois_options *options;

#include <math.h>

static int iteration_no = 1;

static int 
memory_poll_source (akois_db  *db,
                   const char *source_name,
                   const char *source_path)
{
  char   stored_image[256];
  akois_image *image;
  struct stat buf;
  static int logical_clock = 0;

  if (stat (source_path, &buf))
    return -1;

  {
     char fbuf[512];
     struct timeval tv;
     gettimeofday (&tv, 0);
    
     sprintf (fbuf, "%s/%s-%%Y%%m%%d-%%T-%02i.png", MUNIN_STORAGE, source_name, (logical_clock++)%100);
     strftime (stored_image, 256, fbuf, gmtime(&tv.tv_sec));
  }

  {
    char cmd[256];
    sprintf (cmd, "mv '%s' '%s'", source_path, stored_image);
   /* fprintf (stderr, "%s\n", cmd);*/
    system (cmd);
  }

  image = akois_image_new ();
  akois_image_set_path (image, stored_image);
  akois_image_calc_digest (image);
  akois_db_add_image (db, image);
 
  //fprintf (stderr, "a=%s %s\n", source_name, active_stream);
  if (!strcmp (source_name, active_stream) )
    {
      munin_query (db, image, options);
      akois_image_freshen (image); /* this is a call to free any search related
                                      heap allocations, the reason it is here is to
                                      enable keeping some structures alive between
                                      subsequent searches using the same query
                                      image. */
    }

  if (((iteration_no++) % SAVE_INTERVAL) == 0)
    save_db ();
  return 0;
}

#include <glob.h>


static int
memory_iteration (void *data)
{
  char wildcard[512];
  akois_db *db = data;
  glob_t globbuf;
  int i;

  {
  	sprintf (wildcard, "%s/cam0.png", capture_get_dir ());
        memory_poll_source (db, "cam0", wildcard);



	return 1;
  }
  
  sprintf (wildcard, "%s/*.png", capture_get_dir ());
  glob (wildcard, GLOB_PERIOD | GLOB_MARK, NULL, &globbuf);

  for (i=0; i< globbuf.gl_pathc; i++)
    {
      char  buf[255];
      char *ptr;
      char *ptr2;
      strcpy (buf, globbuf.gl_pathv[i]);
      ptr = strrchr (buf, '/');

      if (ptr)
        ptr++;
      else
        ptr = buf;
      ptr2 = strstr (ptr, ".png");
      if (ptr2)
        *ptr2 = '\0';
      //  fprintf (stderr, "\n%s  =>  %s\n", globbuf.gl_pathv[i], ptr);
      memory_poll_source (db, ptr, globbuf.gl_pathv[i]);
    }
  if (!globbuf.gl_pathc)
    usleep (100);
  return 1;
}

static void
parse_args (int    argc,
            char **argv)
{
  char **targ = argv;
  
  targ++; /* ignore program name*/ 
  while (*targ)
    {
      if (!strcmp (*targ, "-w"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-w expects argument\n");
            }
          WINDOW_WIDTH = atoi (*targ);
        }
      else if (!strcmp (*targ, "-h"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-h expects argument\n");
            }
          WINDOW_HEIGHT = atoi (*targ);
        }
      else if (!strcmp (*targ, "-s"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-s expects argument\n");
            }
          SUBDIV = atoi (*targ);
        }
      else if (!strcmp (*targ, "-f"))
        {
          targ++;
          USE_FULLSCREEN = 1;
        }
      else
        {
          if (active_stream)
            free (active_stream);
          active_stream = strdup (*targ);
        }
      targ++;
    }
}


static void
deja_init (void)
{
  options = akois_options_new ();
  akois_options_set_temporal_shadow (options, 100);
}

static void
deja_destroy ()
{
  akois_options_free (options);
}

static akois_db *
memory_init (void)
{
  akois_digest_set_type ("dejab");
  db = akois_db_new ();

  fprintf (stderr, "digest:  %s\n", akois_digest_get_type ());

  { /* check for presence of database, and load it if present */
    FILE *fp;
    fp = fopen ("deja.db", "r");
    if (fp)
      {
        fclose (fp);
        akois_db_load (db, "deja.db");
      }
  }

  /* handler installed after full loading, to avoid overwriting a half loaded
   * database with itself
   */
  signal(SIGINT,  dump_db);
  signal(SIGQUIT, dump_db);

  //akois_db_add_images (db, MUNIN_STORAGE "/*.png");

  return db;
}

static void
memory_destroy ()
{
  dump_db ();
}

static void
video_init (void)
{
  if (SDL_Init (SDL_INIT_VIDEO < 0))
    {
      fprintf(stderr, "SDL_Init failed: %s\n", SDL_GetError ());
      exit (-1);
    }
  
  screen = SDL_SetVideoMode (WINDOW_WIDTH, WINDOW_HEIGHT, SCREEN_BPP, SCREEN_FLAGS);
  SDL_WM_SetCaption ("deja","deja");
  if (!screen)
    {
      fprintf (stderr, "Unable to set SDL mode: %s\n", SDL_GetError ());
      exit (-1);
    }
  SDL_ShowCursor (SDL_DISABLE);
#ifdef USE_OVERLAY
  bmp = SDL_CreateYUVOverlay (CAPTURE_WIDTH, /*CAPTURE_HEIGHT*/ 360, SDL_YUY2_OVERLAY, screen);
#endif
}



static int
event_iteration ()
{
  SDL_Event event;
  while (SDL_PollEvent (&event))
   {
     switch (event.type)
       {
         case SDL_VIDEORESIZE:
           WINDOW_WIDTH  = event.resize.w;
           WINDOW_HEIGHT = event.resize.h;

           screen = SDL_SetVideoMode (WINDOW_WIDTH, WINDOW_HEIGHT, SCREEN_BPP, SCREEN_FLAGS);
           break;
         case SDL_QUIT:
           running=0;
           vm_set_task ("quit");
           break;
         case SDL_KEYUP:
           switch (event.key.keysym.sym)
             {
               case SDLK_1:
                 akois_options_set_temporal_shadow (options, 0);
                 break;
               case SDLK_2:
                 akois_options_set_temporal_shadow (options, 5);
                 break;
               case SDLK_3:
                 akois_options_set_temporal_shadow (options, 10);
                 break;
               case SDLK_4:
                 akois_options_set_temporal_shadow (options, 50);
                 break;
               case SDLK_5:
                 akois_options_set_temporal_shadow (options, 100);
                 break;
               case SDLK_6:
                 akois_options_set_temporal_shadow (options, 200);
                 break;
               case SDLK_7:
                 akois_options_set_temporal_shadow (options, 300);
                 break;
               case SDLK_8:
                 akois_options_set_temporal_shadow (options, 500);
                 break;
               case SDLK_9:
                 akois_options_set_temporal_shadow (options, 1000);
                 break;
               case SDLK_0:
                 akois_options_set_temporal_shadow (options, 2000);
                 break;
               default:
                 running = vm_key_handle (event.key.keysym.sym);
                 break;
             }
           break;
         default:
           break;
       }
   }
  return 1;
}

static void
video_destroy (void)
{
}

#ifndef byteclamp
#define byteclamp(j) do{if(j<0)j=0; else if(j>255)j=255;}while(0)
#endif

#define RGB82YUV8(R,G,B,Y,U,V) \
  do{\
            Y= (( 9798*R + 19234*G + 3736*B)>>15)  +000;\
            U= ((-4817*R - 9470*G + 14320*B)>>15)  +128;\
            V= (( 20152*R - 16875*G - 3277*B)>>15) +128;\
            byteclamp(Y);\
            byteclamp(U);\
            byteclamp(V);\
            }while(0)

static void
munin_display_buf (const unsigned char *pixels,
                   int   width,
                   int   height,
                   int   row_stride,
                   int   x_tile,
                   int   y_tile)
{
  int y;

#ifdef USE_OVERLAY
      for (y = 0; y < 360 /*CAPTURE_HEIGHT*/; y++)
      {
        int       x;
        const unsigned char *src = pixels + row_stride * (y + 16);

        unsigned char *ydst = (unsigned char *) (bmp->pixels[0] + 0 + (y + CAPTURE_HEIGHT * y_tile) * CAPTURE_WIDTH * 2) + CAPTURE_WIDTH * 2 * x_tile;
        unsigned char *udst = ydst + 1;
        unsigned char *vdst = ydst + 3;

        for (x = 0; x < CAPTURE_WIDTH; x++)
          {
            int       R, G, B;
            int       Y, U, V;

            R = src[0];
            G = src[1];
            B = src[2];

            RGB82YUV8 (R, G, B, Y, U, V);

            ydst[0] = Y;
            udst[0] = U;
            vdst[0] = V;

            src += 3;
            ydst += 2;
            if (x % 2)
              {
                udst += 4;
                vdst += 4;
              }
          }
      }
#endif

#ifndef USE_OVERLAY
      for (y = 0; y < CAPTURE_HEIGHT; y++)
      {
        int       x;
        const unsigned char *src = pixels + row_stride * y;
        unsigned char *dst = (unsigned char *) screen->pixels + (y + CAPTURE_HEIGHT * y_tile) * screen->pitch + CAPTURE_WIDTH * x_tile * 4;

        for (x = 0; x < CAPTURE_WIDTH; x++)
          {
	    dst[0] = src[2];
	    dst[1] = src[1];
	    dst[2] = src[0];
	    src+=3;
	    dst+=4;
          }
      }
#endif

}

