#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <math.h>

#include "options.h"
#include "digest.h"

akois_digest * mosaic_digest_new (void);
void mosaic_digest_free (akois_digest *self);
char * mosaic_digest_to_str (akois_digest *self);
akois_digest * mosaic_digest_new_from_str (const char *str);
void mosaic_digest_calc (akois_digest *self, unsigned char *buf, int width, int height, int stride, int x0_px, int y0_px, int width_px, int height_px);
float mosaic_digest_compare (akois_digest  *self, akois_digest  *other, akois_options *options);
void mosaic_digest_freshen (akois_digest *self);
void mosaic_digest_write_png (akois_digest *self, const char   *path);

akois_digest * dejab_digest_new (void);
void dejab_digest_free (akois_digest *self);
char * dejab_digest_to_str (akois_digest *self);
akois_digest * dejab_digest_new_from_str (const char *str);
void dejab_digest_calc (akois_digest *self, unsigned char *buf, int width, int height, int stride, int x0_px, int y0_px, int width_px, int height_px);
float dejab_digest_compare (akois_digest  *self, akois_digest  *other, akois_options *options);
void dejab_digest_freshen (akois_digest *self);
void dejab_digest_write_png (akois_digest *self, const char   *path);


static akois_digest * (*digest_new)          (void) = NULL;
static akois_digest * (*digest_new_from_str) (const char    *str) = NULL;
static void           (*digest_free)         (akois_digest  *self) = NULL;
static char *         (*digest_to_str)       (akois_digest  *self) = NULL;
static void           (*digest_calc)         (akois_digest  *self,
                                              unsigned char *buf,
                                              int            width,
                                              int            height,
                                              int            stride,
                                              int            x0_px,
                                              int            y0_px,
                                              int            width_px,
                                              int            height_px) = NULL;
static float         (*digest_compare)      (akois_digest  *self,
                                              akois_digest  *other,
                                              akois_options *options) = NULL;
static void           (*digest_freshen)      (akois_digest  *self) = NULL;
static void           (*digest_write_png)    (akois_digest  *self,
                                              const char    *path) = NULL;

static char *active_digest = NULL;

void
akois_digest_set_type (const char *new_active_digest)
{
  if (!strcmp (new_active_digest, "mosaic"))
    {
      digest_new          = mosaic_digest_new;
      digest_new_from_str = mosaic_digest_new_from_str;
      digest_free         = mosaic_digest_free;
      digest_to_str       = mosaic_digest_to_str;
      digest_calc         = mosaic_digest_calc;
      digest_compare      = mosaic_digest_compare;
      digest_freshen      = mosaic_digest_freshen;
      digest_write_png    = mosaic_digest_write_png;
    }
  else if (!strcmp (new_active_digest, "dejab"))
    {
      digest_new          = dejab_digest_new;
      digest_new_from_str = dejab_digest_new_from_str;
      digest_free         = dejab_digest_free;
      digest_to_str       = dejab_digest_to_str;
      digest_calc         = dejab_digest_calc;
      digest_compare      = dejab_digest_compare;
      digest_freshen      = dejab_digest_freshen;
      digest_write_png    = dejab_digest_write_png;
    }
  else
    {
      fprintf (stderr, "%s:%i %s(%s) no such digest\n", __FILE__, __LINE__, __FUNCTION__, new_active_digest);
    }

  if (active_digest!=NULL)
    free (active_digest);
  active_digest = strdup (new_active_digest);
}

const char *
akois_digest_get_type (void)
{
  return active_digest;
}


akois_digest *
akois_digest_new (void)
{
  if (digest_new)
    return digest_new ();
  fprintf (stderr, "%s:%i %s() no function set\n", __FILE__, __LINE__, __FUNCTION__);
  return NULL;
}

void
akois_digest_free (akois_digest *self)
{
  if (digest_free)
    digest_free (self);
  else
    fprintf (stderr, "%s:%i %s() no function set\n", __FILE__, __LINE__, __FUNCTION__);
}

char *
akois_digest_to_str (akois_digest *self)
{
  if (digest_to_str)
    return digest_to_str (self);
  fprintf (stderr, "%s:%i %s() no function set\n", __FILE__, __LINE__, __FUNCTION__);
  return NULL;
}

akois_digest *
akois_digest_new_from_str (const char *str)
{
  if (digest_new_from_str)
    return digest_new_from_str (str);
  fprintf (stderr, "%s:%i %s() no function set\n", __FILE__, __LINE__, __FUNCTION__);
  return NULL;
}

void
akois_digest_calc (akois_digest *self,
                   unsigned char *buf,
                   int width,
                   int height,
                   int stride,

                   int x0_px,
                   int y0_px,
                   int width_px,
                   int height_px)
{
  if (digest_calc)
    {
    digest_calc (self, buf, width, height, stride, x0_px, y0_px, width_px, height_px);
    }
  else
    fprintf (stderr, "%s:%i %s() no function set\n", __FILE__, __LINE__, __FUNCTION__);

}


float
akois_digest_compare (akois_digest  *self,
                      akois_digest  *other,
                      akois_options *options)
{
  if (digest_compare)
    return digest_compare (self, other, options);
  fprintf (stderr, "%s:%i %s() no function set\n", __FILE__, __LINE__, __FUNCTION__);
  return 0.0;
}

void
akois_digest_freshen (akois_digest *self)
{
  if (digest_freshen)
    digest_freshen (self);
  else
    fprintf (stderr, "%s:%i %s() no function set\n", __FILE__, __LINE__, __FUNCTION__);
}

void
akois_digest_write_png (akois_digest *self,
                        const char   *path)
{
  if (digest_write_png)
    digest_write_png (self, path);
  else
    fprintf (stderr, "%s:%i %s() no function set\n", __FILE__, __LINE__, __FUNCTION__);
}

extern const char *akois_digest_name;
