/** an akois digest is the middle form of how a almost processed
 *  signature for the image is stored, many different thins can
 *  be put into the digest cathegory, the digest also provides it's
 *  own mean to compare itself to itself, this should allow almost all
 *  approach dependent similarity measures to be kept on this level and
 *  below
 */

#ifndef _AKOIS_DIGEST_H
#define _AKOIS_DIGEST_H

#include "options.h"

typedef struct _akois_digest akois_digest;

void
akois_digest_set_type (const char *digest_type);

const char *
akois_digest_get_type (void);

/* create a new empty digest object
 */
akois_digest *
akois_digest_new (void);

/* free the data digest and all related memory
 */
void
akois_digest_free (akois_digest *self);

/* serialize data to string (NOT NEEDED FOR DISPLAY)
 */
char *
akois_digest_to_str (akois_digest *self);

/* parse a digest from a single line string
 */
akois_digest *
akois_digest_new_from_str (const char *str);

/* create digest based on external image data contained in
 * "path"
 */
void
akois_digest_calc (akois_digest *self,
                   unsigned char *buf,
                   int width,
                   int height,
                   int stride,

                   int x0_px,
                   int y0_px,
                   int width_px,
                   int height_px);


/* compare this digest to another digest, can
 * use intermediate caching data structures.
 */
float
akois_digest_compare (akois_digest  *self,
                      akois_digest  *other,
                      akois_options *options);

/* cleanup any cached version of "self"
 * should be called when one is done using a
 * digests compare method to avoid memory leaks of
 * cached representations.
 */
void
akois_digest_freshen (akois_digest *self);

/* write a visualization of the digest as
 * a png file.*/
void
akois_digest_write_png (akois_digest *self,
                        const char   *path);


#endif /* _AKOIS_DIGEST_H */
