#include "dispatch.h"
#include "capture.h"
#include "lost-in-time.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <SDL.h>

static double     threshold      = 0.8;

static char *MONITOR_DIRd = NULL;
static int   dump_mask = 0;

void capture_set_mask (int do_dump_mask)
{
	dump_mask = do_dump_mask;
}

void capture_set_dir (const char *dir)
{
  if (MONITOR_DIRd)
	  free (MONITOR_DIRd);
  MONITOR_DIRd = strdup (dir);
  { /* create directory */
	char buf[256];
	sprintf (buf, "install -d %s", dir);
	system (buf);
  }
}
const char *capture_get_dir (void)
{
  if (MONITOR_DIRd)
	return MONITOR_DIRd;
  return "/tmp/";
}

static void parse_args (int    argc, char **argv);
static char *active_stream = NULL;


static int  event_iteration ();
static int running=1;

static int
eye_block (void *data)
{
  struct stat buf;
  char sourcepath[512];

  sprintf (sourcepath, "%s%s.png", capture_get_dir (), active_stream);

  while (!stat (sourcepath, &buf))
  {
    usleep (100);
  }
  return 1;
}

int
munin_eye (int argc,
           char **argv)
{
  Dispatch *dispatch;

  parse_args (argc, argv);
  dispatch = dispatch_new ();

  dispatch_add_task (dispatch, 0, NULL, eye_block, NULL, NULL);
  dispatch_add_task (dispatch, 0, NULL, event_iteration, NULL, NULL);
  dispatch_add_task (dispatch, 0, capture_init, capture_iteration, capture_destroy, NULL);

  running = 1;
  while (dispatch_iterate (dispatch) && running);

  dispatch_destroy (dispatch);
  return 0;
}

static int
event_iteration ()
{
  SDL_Event event;
  while (SDL_PollEvent (&event))
   {
     switch (event.type)
       {
         case SDL_QUIT:
           running=0;
           vm_set_task ("quit");
           break;
         case SDL_KEYUP:
           switch (event.key.keysym.sym)
             {
               /*FIXME: handle setting threshold */
               default:
                 running = vm_key_handle (event.key.keysym.sym);
                 break;
             }
           break;
         default:
           break;
       }
   }
  return 1;
}


#include <oxide/oxide.h>
static Oxide *oxide = NULL;
static Oxide *scene_detector = NULL;

int CAPTURE_WIDTH=128;
int CAPTURE_HEIGHT=96;

#include <gggl/image/gggl_image.h>

static GgglImage *temp_image = NULL;
static GgglImage *mask_image = NULL;

#include <math.h>

int mask_capture_iteration ()
{
  double    difference;

  oxide_process (oxide);
  difference = oxide_get_property_double (scene_detector, "difference");

  if (difference > threshold)
    {
      {
      Oxide *oxide;
      Oxide *memsrc;
      Oxide *png_save;

      char *xml_src =
      "<oxide>"
      "  <op id='png_save' type='png_save'>"
      "    <property name='compression'>0</property>"
      "  </op>"
      "  <op id='memsrc' type='memsrc'>"
      "  </op>"
      "</oxide>";
            
      char      buf[100];
      sprintf (buf, "%p", mask_image);
         
      oxide = oxide_parse_xml_chunk (NULL, xml_src);
          
      memsrc = oxide_get_element_by_id (oxide, "memsrc"); 
      png_save = oxide_get_element_by_id (oxide, "png_save");
      oxide_set_property (memsrc, "gggl_image", buf);

      sprintf (buf, "%s%s.png-", capture_get_dir (), "mask");
      oxide_set_property (png_save, "resource", buf);

      oxide_process (oxide);
      oxide_free (oxide);

      {
        char cmdbuf[512];

        sprintf (cmdbuf, "mv -f %s%s.png- %s%s.png", capture_get_dir(), "mask", capture_get_dir (), "mask");
        system (cmdbuf);
      }
      }


      Oxide *oxide;
      Oxide *memsrc;
      Oxide *png_save;

      char *xml_src =
      "<oxide>"
      "  <op id='png_save' type='png_save'>"
      "    <property name='compression'>0</property>"
      "  </op>"
      "  <op type='hflip'/>"
      "  <op id='memsrc' type='memsrc'>"
      "  </op>"
      "</oxide>";
            
      char      buf[100];
      sprintf (buf, "%p", temp_image);
         
      oxide = oxide_parse_xml_chunk (NULL, xml_src);
          
      memsrc = oxide_get_element_by_id (oxide, "memsrc"); 
      png_save = oxide_get_element_by_id (oxide, "png_save");
      oxide_set_property (memsrc, "gggl_image", buf);

      sprintf (buf, "%s%s.png-", capture_get_dir (), active_stream);
      oxide_set_property (png_save, "resource", buf);

      oxide_process (oxide);
      oxide_free (oxide);

      {
        char cmdbuf[512];

        sprintf (cmdbuf, "mv -f %s%s.png- %s%s.png", capture_get_dir(), active_stream, capture_get_dir (), active_stream);
        system (cmdbuf);
      }
   }
  else
  {
	usleep (150000);
  }

  return 1;
}


int capture_iteration ()
{
  double    difference;

  if (dump_mask)
	  return mask_capture_iteration ();
  oxide_process (oxide);
  difference = oxide_get_property_double (scene_detector, "difference");

  if (difference > threshold)
    {
      Oxide *oxide;
      Oxide *memsrc;
      Oxide *png_save;

      char *xml_src =
      "<oxide>"
      "  <op id='png_save' type='png_save'>"
      "    <property name='compression'>0</property>"
      "  </op>"
      "  <op type='hflip'/>"
      "  <op id='memsrc' type='memsrc'>"
      "  </op>"
      "</oxide>";
            
      char      buf[100];
      sprintf (buf, "%p", temp_image);
         
      oxide = oxide_parse_xml_chunk (NULL, xml_src);
          
      memsrc = oxide_get_element_by_id (oxide, "memsrc"); 
      png_save = oxide_get_element_by_id (oxide, "png_save");
      oxide_set_property (memsrc, "gggl_image", buf);

      sprintf (buf, "%s%s.png-", capture_get_dir (), active_stream);
      oxide_set_property (png_save, "resource", buf);

      oxide_process (oxide);
      oxide_free (oxide);

      {
        char cmdbuf[512];

        sprintf (cmdbuf, "mv -f %s%s.png- %s%s.png", capture_get_dir(), active_stream, capture_get_dir (), active_stream);
        system (cmdbuf);
      }
   }

  return 1;
}


static void
parse_args (int    argc,
            char **argv)
{
  char **targ = argv;
  
  targ++; /* ignore program name*/ 
  while (*targ)
    {
      if (!strcmp (*targ, "-t"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-t expects argument\n");
            }
          threshold = atof (*targ);
        }
      else if (!strcmp (*targ, "-w"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-w expects argument\n");
            }
          CAPTURE_WIDTH = atoi (*targ);
        }
      else if (!strcmp (*targ, "-h"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-h expects argument\n");
            }
          CAPTURE_HEIGHT = atoi (*targ);
        }
      else
        {
          if (active_stream)
            free (active_stream);
          active_stream = strdup (*targ);
        }
      targ++;
    }
}

/* eye */

#include <gggl/gggl.h>          /* for loading additional op dir */

/* global variables are bad, but an easy way to implement
 * command line arguments
 */

#define DEBUG_MODE
#define MBLUR_STR_EXHIBIT "0.01"
#define MBLUR_STR_MOTION "0.6"

#define   MBLUR_STR MBLUR_STR_MOTION

static void mask_capture_init (void)
{
  if (!active_stream)
    active_stream = strdup ("cam0");

   oxide = oxide_parse_xml_chunk (NULL, "<oxide>"
      "  <op id='memdst' type='memdst2' />"
      "  <op type='clone' idref='1'/>"
      "  <op type='display'/>"
      "  <op type='clone' idref='43'/>"
      "  <op id='maskdst' type='memdst2' />"
      "  <op type='median'>"
      "    <output id='43'/>"
      "  </op>"
      "    <op type='key_clean_slate' level='0.02' smooth_range='0.00'>"
      "       <op type='mblur' opacity='0.001'/>"
      "       <op type='clone' idref='42'/>"
      "    </op>"
      "  <op type='scale'>"
      "    <property name='width_px'>32</property>"
      "    <property name='height_px'>32</property>"
      "    <output id='42'/>"
      "  </op>"
      "  <op type='boxblur' horizontal='20' vertical='20'>"
      "  </op>"
      
      "  <op type='clone' idref='1'/>"

      "  <op id='scene_detector' type='avgdiff'>"
      "    <op id='memory' type='mblur'>"
      "    <output id='3'/>"
      "      <property name='opacity'>" MBLUR_STR  "</property>"
      "    </op>"
      "    <op type='clone'>"
      "      <property name='idref'>2</property>"
      "    </op>"
      "  </op>"
      "  <op type='scale'>"
      "    <property name='width_px'>16</property>"
      "    <property name='height_px'>12</property>"
      "    <output id='2'/>"
      "  </op>"
      "  <op id='Scaler' type='scale'>"
      "    <property name='width_px'>128</property>"
      "    <property name='height_px'>96</property>"
      "    <output id='1'/>"
      "  </op>"
      "  <op id='Depad' type='depad'>"
      "    <property name='left'>0</property>"
      "    <property name='right'>0</property>"
      "    <property name='top'>0</property>"
      "    <property name='bottom'>0</property>"
      "  </op>"
      "  <op id='Source' type='v4l'>"
      "    <property name='width'>320</property>"
      "    <property name='height'>240</property>"
      "  </op>"
      "</oxide>");

  temp_image = gggl_image_alloc (gggl_pixfmt ("rgb8"), CAPTURE_WIDTH, CAPTURE_HEIGHT); 
  mask_image = gggl_image_alloc (gggl_pixfmt ("rgb8"), 32, 32);

  if (!temp_image)
    {
      fprintf (stderr, "unable to allocate temp image\n");
      exit (-1);
    }
  {
    if (getenv ("DEPAD_LEFT"))
       oxide_set_property_int (oxide_get_element_by_id (oxide, "Depad"), "left", atoi(getenv("DEPAD_LEFT")));
    if (getenv ("DEPAD_RIGHT"))
       oxide_set_property_int (oxide_get_element_by_id (oxide, "Depad"), "right", atoi(getenv("DEPAD_RIGHT")));
    if (getenv ("DEPAD_TOP"))
       oxide_set_property_int (oxide_get_element_by_id (oxide, "Depad"), "top", atoi(getenv("DEPAD_TOP")));
    if (getenv ("DEPAD_BOTTOM"))
       oxide_set_property_int (oxide_get_element_by_id (oxide, "Depad"), "bottom", atoi(getenv("DEPAD_BOTTOM")));
  }
  {
    char buf[256];
    sprintf (buf, "%p", temp_image);
    oxide_set_property (oxide_get_element_by_id (oxide, "memdst"), "gggl_image", buf);

    sprintf (buf, "%p", mask_image);
    oxide_set_property (oxide_get_element_by_id (oxide, "maskdst"), "gggl_image", buf);

    if (CAPTURE_WIDTH > 320)
    {
      oxide_set_property_int (oxide_get_element_by_id (oxide, "Source"), "width", CAPTURE_WIDTH);
      oxide_set_property_int (oxide_get_element_by_id (oxide, "Source"), "height", CAPTURE_HEIGHT);
    }
    oxide_set_property_int (oxide_get_element_by_id (oxide, "Scaler"), "width_px", CAPTURE_WIDTH);
    oxide_set_property_int (oxide_get_element_by_id (oxide, "Scaler"), "height_px", CAPTURE_HEIGHT);
    scene_detector = oxide_get_element_by_id (oxide, "scene_detector");
  }
}

void
capture_init (void)
{
  if (dump_mask)
  {
	  mask_capture_init ();
	  return;
  }
  if (!active_stream)
    active_stream = strdup ("cam0");

   oxide = oxide_parse_xml_chunk (NULL, "<oxide>"
      "  <op id='memdst' type='memdst2' />"
      "  <op type='clone' idref='1'/>"/*
      "  <op type='display-sdl' title='eye'/>"
      "  <op type='clone' idref='1'/>"*/
#ifdef DEBUG_MODE___
      "  <op type='display'/>"
      "  <op type='clone' idref='3'/>"
      "  <op type='display'/>"
      "  <op type='clone' idref='2'/>"
#endif
      "  <op id='scene_detector' type='avgdiff'>"
      "    <op id='memory' type='mblur'>"
      "    <output id='3'/>"
      "      <property name='opacity'>" MBLUR_STR  "</property>"
      "    </op>"
      "    <op type='clone'>"
      "      <property name='idref'>2</property>"
      "    </op>"
      "  </op>"
      "  <op type='scale'>"
      "    <property name='width_px'>16</property>"
      "    <property name='height_px'>12</property>"
      "    <output id='2'/>"
      "  </op>"
      "  <op id='Scaler' type='scale'>"
      "    <property name='width_px'>128</property>"
      "    <property name='height_px'>96</property>"
      "    <output id='1'/>"
      "  </op>"
      "  <op id='Depad' type='depad'>"
      "    <property name='left'>0</property>"
      "    <property name='right'>0</property>"
      "    <property name='top'>0</property>"
      "    <property name='bottom'>0</property>"
      "  </op>"
      "  <op id='Source' type='v4l'>"
      "    <property name='width'>320</property>"
      "    <property name='height'>240</property>"
      "  </op>"
      "</oxide>");

  temp_image = gggl_image_alloc (gggl_pixfmt ("rgb8"), CAPTURE_WIDTH, CAPTURE_HEIGHT); 

  if (!temp_image)
    {
      fprintf (stderr, "unable to allocate temp image\n");
      exit (-1);
    }
  {
    char buf[256];
    sprintf (buf, "%p", temp_image);
    oxide_set_property (oxide_get_element_by_id (oxide, "memdst"), "gggl_image", buf);


  {
    if (getenv ("DEPAD_LEFT"))
       oxide_set_property_int (oxide_get_element_by_id (oxide, "Depad"), "left", atoi(getenv("DEPAD_LEFT")));
    if (getenv ("DEPAD_RIGHT"))
       oxide_set_property_int (oxide_get_element_by_id (oxide, "Depad"), "right", atoi(getenv("DEPAD_RIGHT")));
    if (getenv ("DEPAD_TOP"))
       oxide_set_property_int (oxide_get_element_by_id (oxide, "Depad"), "top", atoi(getenv("DEPAD_TOP")));
    if (getenv ("DEPAD_BOTTOM"))
       oxide_set_property_int (oxide_get_element_by_id (oxide, "Depad"), "bottom", atoi(getenv("DEPAD_BOTTOM")));
  }

    if (CAPTURE_WIDTH > 320)
    {
      oxide_set_property_int (oxide_get_element_by_id (oxide, "Source"), "width", CAPTURE_WIDTH);
      oxide_set_property_int (oxide_get_element_by_id (oxide, "Source"), "height", CAPTURE_HEIGHT);
    }
    oxide_set_property_int (oxide_get_element_by_id (oxide, "Scaler"), "width_px", CAPTURE_WIDTH);
    oxide_set_property_int (oxide_get_element_by_id (oxide, "Scaler"), "height_px", CAPTURE_HEIGHT);
    scene_detector = oxide_get_element_by_id (oxide, "scene_detector");
  }
}

void
capture_destroy ()
{
  free (temp_image);
  oxide_free (oxide);
}

