#ifndef _AKOIS_IMAGE_H
#define _AKOIS_IMAGE_H

#include "options.h"

typedef struct _akois_image  akois_image;

/* object construction, destruction, serialization and parsing
 */

/* allocate a new image object */
akois_image *
akois_image_new (void);

/* free image object and related data*/
void
akois_image_free (akois_image *self);

/* allocate a new image object from file, the file has the following
format:

<path_to_image> <digest,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,>

the image path may not contain spaces, a single space delimits the image
name from the following string, which is a serialized form of the image
digest/search signature, the signature is terminated the char before  the first
newline.
*/
akois_image *
akois_image_new_from_str (const char *str);

/*
 * serialize image and containing digest to a string
 */
char *
akois_image_to_str (akois_image *self);

/******************************/

/* set the path of the associated image */
void
akois_image_set_path (akois_image *self,
                      const char  *path);

const char *
akois_image_get_path (akois_image *self);

/* make the image calculate it's own digest */
void
akois_image_calc_digest (akois_image *self);

/* An additional query image feature, originally developed
 * for subimage search.
 */
void
akois_image_calc_sub_digest (akois_image *self,
                             double       x0,
                             double       y0,
                             double       width,  
                             double       height);
                                   
/* provide the distance between two images,
 * uses digest comparison internally
 */
double
akois_image_distance (akois_image   *self,
                      akois_image   *other,
                      akois_options *options);

/* free up intermediate memory allocated during
 * comparisons, only needs to be freed after last comparison
 * this image makes towards others
 */
void
akois_image_freshen (akois_image *self);


/* write a png image of the digest
 */
void
akois_image_write_png (akois_image *self,
                       const char  *path);


void
akois_image_adopt_sub_digest (akois_image *self,
                              akois_image *other);


int  
akois_image_get_hits (akois_image *self);

void
akois_image_set_hits (akois_image *self,
                       int          hits);
void
akois_image_inc_hits (akois_image *self);

int  
akois_image_get_compares (akois_image *self);

void
akois_image_set_compares (akois_image *self,
                          int          compares);

void
akois_image_inc_compares (akois_image *self);

int
akois_image_get_data (akois_image    *self,
                      int            *width,
                      int            *height,
                      int            *row_stride,
                      unsigned char **pixels);

#endif /* _AKOIS_IMAGE_H */
