// 1024
//  512
// 1536
#define DB_SIZE       1024
#define DB_SLACK      512
#define SAVE_INTERVAL 1000

static int   SUBDIV         = 12;

// ABCDEF0123
//
// W64

#include "dispatch.h"
/*
Setup mode / surveilannce combined cam
  
Adds new sources as new sources register with the image db.
(Good debug would be to write the filenames using OSD
*/

#define MUNIN_STORAGE "mosaic_images"


#include <png.h>
#include <stdlib.h>

int
save_png (const char *path,
          int         width,
          int         height,
          int         row_stride,
          unsigned char      *data)
{
  FILE          *fp;
  int           i;
  png_struct    *png;
  png_info      *info;
  png_byte     **rows;
  png_color_16   white;

  fp = fopen (path, "wb");
  if (!fp)
    {
      return -1;
    }
  
  rows = malloc (height * sizeof (png_byte*));
  if (rows == NULL)
    return -1;

  for (i=0; i<height; i++)
    rows[i] = (png_byte *) data + i * row_stride;

  png = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (png == NULL)
    return -1;

  info = png_create_info_struct (png);

  if (setjmp (png_jmpbuf (png)))
    return -1;
  png_set_compression_level (png, 3);
  png_init_io (png, fp);

  png_set_IHDR (png, info,
     width, height, 8, PNG_COLOR_TYPE_RGB_ALPHA,
     PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_DEFAULT);

  white.red = 0xff;
  white.blue = 0xff;
  white.green = 0xff;
  png_set_bKGD (png, info, &white);

  png_write_info (png, info);
  png_write_image (png, rows);
  png_write_end (png, info);

  png_destroy_write_struct (&png, &info);
  free (rows);
  fclose (fp);

  return 0;
}

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include "db.h"
#include "load_image.h"
#include "digest.h"
#include "capture.h"
#include "lost-in-time.h"
#include <SDL.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>

extern int CAPTURE_WIDTH;
extern int CAPTURE_HEIGHT;

static int   USE_FULLSCREEN = 0;
static int   WINDOW_WIDTH   = 1024;
static int   WINDOW_HEIGHT  = 768;

static unsigned char *fb = NULL;

static void parse_args (int    argc, char **argv);
static char *active_stream = NULL;

static int  event_iteration ();
static int  memory_iteration (void *db);
static akois_db * memory_init (void);
static void memory_destroy ();
static void deja_init ();
static void deja_destroy ();
static void video_init ();
static void video_destroy ();
static void last_digest_init ();
static void last_digest_destroy ();

static int running=1;

static int debug=0;


int
munin_mosaic (int argc,
              char **argv)
{
  Dispatch *dispatch;
  akois_db *db;

  CAPTURE_WIDTH = 84;
  CAPTURE_HEIGHT = 63;

  parse_args (argc, argv);
  if (!active_stream)
    active_stream = strdup ("cam0");

  dispatch = dispatch_new ();
  db=memory_init ();

  {
	  char buf[256];
	  sprintf (buf, "install -d %s", MUNIN_STORAGE);
	  system (buf);
  }
  capture_set_dir ("/tmp/munin_mosaic/");
  dispatch_add_task (dispatch, 0, capture_init, capture_iteration,  capture_destroy,  NULL);
  dispatch_add_task (dispatch, 0, last_digest_init,      NULL,  last_digest_destroy,  NULL);
  dispatch_add_task (dispatch, 0, video_init, event_iteration,  video_destroy,  NULL);
  dispatch_add_task (dispatch, 0, NULL,       memory_iteration, memory_destroy, db);
  dispatch_add_task (dispatch, 0, deja_init,  NULL,             deja_destroy,   NULL);

  fb = malloc (CAPTURE_WIDTH * CAPTURE_HEIGHT * 4 * 12 * 12);
  
  running = 1;
  while (dispatch_iterate (dispatch) && running);

  dispatch_destroy (dispatch);

  free (fb);

  return 0;
}


#include <oxide/oxide.h>

#define hor_tiles SUBDIV
#define ver_tiles SUBDIV

#define SCREEN_FLAGS   (SDL_RESIZABLE | SDL_HWSURFACE | (USE_FULLSCREEN?SDL_FULLSCREEN:0))


#define BUFFER_HEIGHT  (CAPTURE_HEIGHT*ver_tiles)
#define BUFFER_WIDTH   (CAPTURE_WIDTH*hor_tiles)


#define SCREEN_BPP     32

#include <gggl/image/gggl_image.h>

/* Need OSD features,
 * allow ui to be slow
 */

static SDL_Surface *screen = NULL;
static SDL_Overlay *bmp = NULL;

static akois_db *db;  /* to have it available for the SIGINT proc */

static void
munin_display_buf (const unsigned char *pixels,
                   int   width,
                   int   height,
                   int   row_stride,
                   int   tile_x,
                   int   tile_y);

static void save_db()
{
  system ("cp mosaic.db mosaic.db~");
  akois_db_save (db, "mosaic.db");
}

static void dump_db()
{
  fprintf (stderr, "\nMunin is terminating\n");
  SDL_Quit ();
  save_db ();
  akois_db_free (db);
  //exit (0);
}

static void
munin_display (akois_image *image,
               int         tile_x,
               int         tile_y)
{
  unsigned char *pixels;
  int width;
  int height;
  int row_stride;
  //const char *file = akois_image_get_path (image);

  if (!akois_image_get_data (image, &width, &height, &row_stride, &pixels))
    {
      munin_display_buf (pixels, width, height, row_stride,
                         tile_x, tile_y);
      //free (pixels);
    }
}

static akois_image * last_digest[32*32];

static void last_digest_init ()
{
  int i;
  for (i=0; i<32*32;i++)
    last_digest[i]=akois_image_new ();
}

static void last_digest_destroy ()
{
  int i;
  for (i=0; i<32*32;i++)
    akois_image_free (last_digest[i]);
}

#if 0
/* mosaic query */
static void
old_munin_query (akois_db      *db,
                 akois_image   *image,
                 akois_options *options)
{
  int x,y;

  int nops=0;
  int queries=0;

  for (y=0;y<ver_tiles;y++)
    {
  SDL_LockYUVOverlay (bmp);
    for (x=0;x<hor_tiles;x++)
      {
         if (!running)
           return;
      
         akois_image_calc_sub_digest (image, 
                                      (float)x/hor_tiles,
                                      (float)y/ver_tiles,
                                      1.0/hor_tiles,
                                      1.0/ver_tiles);
         
         /* Check if replacement really is neccesary  */
         if (akois_image_distance(image, last_digest[y*32+x], options) > 0.003)
           {
             akois_image *result = akois_db_best (db, image, options);
             if (result)
                munin_display (akois_image_get_path (result), x, y);
             akois_image_adopt_sub_digest (last_digest[y*32+x], image);
             queries++;
           }
         else
           {
             nops++;
           }
         
      }
  SDL_UnlockYUVOverlay (bmp);

         event_iteration ();
    }
  {
    SDL_Rect  rect;

    rect.x = 0;
    rect.y = 0;
    rect.w = WINDOW_WIDTH;
    rect.h = WINDOW_HEIGHT;

    SDL_DisplayYUVOverlay (bmp, &rect);

  }
}
#endif

static void
munin_query (akois_db      *db,
             akois_image   *image,
             akois_options *options)
{
  int got_one=0;
  int x,y, i;
  akois_image *samples[hor_tiles*ver_tiles];
  akois_image *results[hor_tiles*ver_tiles];

  
  i=0;
  for (y=0;y<ver_tiles;y++)
    for (x=0;x<hor_tiles;x++)
      {
         akois_image_calc_sub_digest (image, 
                                      (float)x/hor_tiles,
                                      (float)y/ver_tiles,
                                      1.0/hor_tiles,
                                      1.0/ver_tiles);
         if (akois_image_distance(image, last_digest[y*32+x], options) > 0.003)
           {
              akois_image_adopt_sub_digest (last_digest[y*32+x], image);
              samples[i]=last_digest[y*32+x];
           }
         else
           {
              samples[i]=NULL;
           }
         results[i]=NULL;
         i++;
      }
  i=0;
  akois_db_best_many (db, options, samples, results, hor_tiles*ver_tiles);  
  SDL_LockYUVOverlay (bmp);
  for (y=0;y<ver_tiles;y++)
    for (x=0;x<hor_tiles;x++)
      {
         if (results[i])
	 {
           munin_display (results[i], x,y);
	   got_one=1;
	 }
         i++;
      }
  SDL_UnlockYUVOverlay (bmp);
  {
    SDL_Rect  rect;

    rect.x = 0;
    rect.y = 0;
    rect.w = WINDOW_WIDTH;
    rect.h = WINDOW_HEIGHT;

    SDL_DisplayYUVOverlay (bmp, &rect);
  }
  if (0 && got_one){
	static int frame_no=0;
	char buf[256];
	sprintf (buf, "/home/pippin/mosaic_render/%08i.png", frame_no++);
	save_png (buf, BUFFER_WIDTH, BUFFER_HEIGHT, BUFFER_WIDTH * 4, fb);
  }
}

akois_options *options;

#include <math.h>

static int 
memory_poll_source (akois_db  *db,
                   const char *source_name,
                   const char *source_path)
{
  char   stored_image[256];
  akois_image *image;
  struct stat buf;
  static int logical_clock = 0;

  if (stat (source_path, &buf))
    return -1;

  {
     char format_buf[512];
     struct timeval tv;
     gettimeofday (&tv, 0);
    
     sprintf (format_buf, "%s/%s-%%Y%%m%%d-%%T-%02i.png", MUNIN_STORAGE, source_name, (logical_clock++)%100);
     strftime (stored_image, 256, format_buf, gmtime(&tv.tv_sec));
  }

  {
    char cmd[256];
    sprintf (cmd, "mv '%s' '%s'", source_path, stored_image);
    system (cmd);
  }

  image = akois_image_new ();
  akois_image_set_path (image, stored_image);
  akois_image_calc_digest (image);
  akois_db_add_image (db, image);
 
  //fprintf (stderr, "a=%s %s\n", source_name, active_stream);
  if (bmp && !debug && !strcmp (source_name, active_stream) )
    {
      //fprintf (stderr, "munin_query (db, image, options);\n");
      munin_query (db, image, options);
      akois_image_freshen (image); /* this is a call to free any search related
                                      heap allocations, the reason it is here is to
                                      enable keeping some structures alive between
                                      subsequent searches using the same query
                                      image. */
    }

  return 0;
}

#include <glob.h>

static int iteration_no=1;

static int
memory_iteration (void *data)
{
  char wildcard[512];
  akois_db *db = data;
  glob_t globbuf;
  int i;
  sprintf (wildcard, "%s/*.png", capture_get_dir ());
  glob (wildcard, GLOB_PERIOD | GLOB_MARK, NULL, &globbuf);

  for (i=0; i< globbuf.gl_pathc; i++)
    {
      char  buf[255];
      char *ptr;
      char *ptr2;
      strcpy (buf, globbuf.gl_pathv[i]);
      ptr = strrchr (buf, '/');

      if (ptr)
        ptr++;
      else
        ptr = buf;
      ptr2 = strstr (ptr, ".png");
      if (ptr2)
        *ptr2 = '\0';
      //  fprintf (stderr, "\n%s  =>  %s\n", globbuf.gl_pathv[i], ptr);
      memory_poll_source (db, ptr, globbuf.gl_pathv[i]);
    }
  if (!globbuf.gl_pathc)
    usleep (100);

  akois_db_reduce_to (db, DB_SIZE, DB_SLACK);

  iteration_no++;
  if (((iteration_no++) % SAVE_INTERVAL) == 0)
     save_db ();

  globfree (&globbuf);
  
  return 1;
}

static void
parse_args (int    argc,
            char **argv)
{
  char **targ = argv;
  
  targ++; /* ignore program name*/ 
  while (*targ)
    {
      if (!strcmp (*targ, "-w"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-w expects argument\n");
            }
          WINDOW_WIDTH = atoi (*targ);
        }
      else if (!strcmp (*targ, "-h"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-h expects argument\n");
            }
          WINDOW_HEIGHT = atoi (*targ);
        }
      else if (!strcmp (*targ, "-s"))
        {
          targ++;
          if (!*targ)
            {
              fprintf (stderr, "-s expects argument\n");
            }
          SUBDIV = atoi (*targ);
        }
      else if (!strcmp (*targ, "-f"))
        {
          USE_FULLSCREEN = 1;
        }
      else
        {
          if (active_stream)
            free (active_stream);
          active_stream = strdup (*targ);
        }
      targ++;
    }
}

static void
deja_init (void)
{
  options = akois_options_new ();
  akois_options_set_temporal_shadow (options, 0);

  akois_options_set_complexity (options, 3);
  akois_options_set_luminance (options, 0.6);
  akois_options_set_chrominance (options, 1);
}

static void
deja_destroy ()
{
  akois_options_free (options);
}

static akois_db *
memory_init (void)
{
  akois_digest_set_type ("mosaic");

  db = akois_db_new ();

  fprintf (stderr, "digest:  %s\n", akois_digest_get_type ());

  { /* check for presence of database, and load it if present */
    FILE *fp;
    fp = fopen ("mosaic.db", "r");
    if (fp)
      {
        fclose (fp);
        akois_db_load (db, "mosaic.db");
      }
  }

  /* handler installed after full loading, to avoid overwriting a half loaded
   * database with itself
   */
  signal(SIGINT,  dump_db);
  signal(SIGQUIT, dump_db);

  //akois_db_add_images (db, MUNIN_STORAGE "/*.png");

  return db;
}

static void
memory_destroy ()
{
  dump_db ();
}

static void
video_init (void)
{
  if (SDL_Init (SDL_INIT_VIDEO < 0))
    {
      fprintf(stderr, "SDL_Init failed: %s\n", SDL_GetError ());
      exit (-1);
    }
  
  screen = SDL_SetVideoMode (WINDOW_WIDTH, WINDOW_HEIGHT, SCREEN_BPP, SCREEN_FLAGS);
  if (!screen)
    {
      fprintf (stderr, "Unable to set SDL mode: %s\n", SDL_GetError ());
      exit (-1);
    }
  SDL_WM_SetCaption ("mosaic","mosaic");
  SDL_ShowCursor (SDL_DISABLE);
  bmp = SDL_CreateYUVOverlay (BUFFER_WIDTH, BUFFER_HEIGHT, SDL_YUY2_OVERLAY, screen);
}


void     babl_fish_stats                (FILE           *file);

static int
event_iteration ()
{
  SDL_Event event;
  while (SDL_PollEvent (&event))
   {
     switch (event.type)
       {
         case SDL_VIDEORESIZE:
           WINDOW_WIDTH  = event.resize.w;
           WINDOW_HEIGHT = event.resize.h;

           screen = SDL_SetVideoMode (WINDOW_WIDTH, WINDOW_HEIGHT, SCREEN_BPP, SCREEN_FLAGS);
           break;
         case SDL_QUIT:
           vm_set_task ("quit");
           running = 0;
           break;
         case SDL_KEYUP:
           switch (event.key.keysym.sym)
             {
               case SDLK_d: /* debug babl (memory handling got out of hand at some point) */
                  {
                     char logfile_name[]="/tmp/babl-stats.html";
                     FILE *logfile;
                     logfile = fopen (logfile_name, "w");
                     if (logfile)
                       {
                        babl_fish_stats (logfile);
                        fclose (logfile);
                       }
                  }
                 break;
               case SDLK_a:
                 SUBDIV++;
                 if (SUBDIV>63)
                   SUBDIV=63;
                 SDL_FreeYUVOverlay (bmp); 
                 bmp = SDL_CreateYUVOverlay (BUFFER_WIDTH, BUFFER_HEIGHT, SDL_YUY2_OVERLAY, screen);
                 break;
               case SDLK_z:
                 SUBDIV--;
                 if (SUBDIV<1)
                   SUBDIV=1;
                 SDL_FreeYUVOverlay (bmp); 
                 bmp = SDL_CreateYUVOverlay (BUFFER_WIDTH, BUFFER_HEIGHT, SDL_YUY2_OVERLAY, screen);
                 break;
               case SDLK_1:
                 akois_options_set_temporal_shadow (options, 0);
                 break;
               case SDLK_2:
                 akois_options_set_temporal_shadow (options, 5);
                 break;
               case SDLK_3:
                 akois_options_set_temporal_shadow (options, 10);
                 break;
               case SDLK_4:
                 akois_options_set_temporal_shadow (options, 50);
                 break;
               case SDLK_5:
                 akois_options_set_temporal_shadow (options, 100);
                 break;
               case SDLK_6:
                 akois_options_set_temporal_shadow (options, 200);
                 break;
               case SDLK_7:
                 akois_options_set_temporal_shadow (options, 300);
                 break;
               case SDLK_8:
                 akois_options_set_temporal_shadow (options, 500);
                 break;
               case SDLK_9:
                 akois_options_set_temporal_shadow (options, 1000);
                 break;
               case SDLK_0:
                 akois_options_set_temporal_shadow (options, 2000);
                 break;
               default:
                 running = vm_key_handle (event.key.keysym.sym);
                 break;
             }
           break;
         default:
           break;
       }
   }
  return 1;
}


static void
video_destroy (void)
{
}



#ifndef byteclamp
#define byteclamp(j) do{if(j<0)j=0; else if(j>255)j=255;}while(0)
#endif

#define RGB82YUV8(R,G,B,Y,U,V) \
  do{\
            Y= (( 9798*R + 19234*G + 3736*B)>>15)  +000;\
            U= ((-4817*R - 9470*G + 14320*B)>>15)  +128;\
            V= (( 20152*R - 16875*G - 3277*B)>>15) +128;\
            byteclamp(Y);\
            byteclamp(U);\
            byteclamp(V);\
            }while(0)

static void
munin_display_buf (const unsigned char *pixels,
                   int   width,
                   int   height,
                   int   row_stride,
                   int   x_tile,
                   int   y_tile)
{
  int y;
      for (y = 0; y < CAPTURE_HEIGHT; y++)
      {
        int       x;
        const unsigned char *src = pixels + row_stride * y;

        unsigned char *ydst = (unsigned char *) (bmp->pixels[0] + 0 + (y + CAPTURE_HEIGHT * y_tile) * BUFFER_WIDTH * 2) + CAPTURE_WIDTH * 2 * x_tile;
        unsigned char *udst = ydst + 1;
        unsigned char *vdst = ydst + 3;

        for (x = 0; x < CAPTURE_WIDTH; x++)
          {
            int       R, G, B;
            int       Y, U, V;

            R = src[0];
            G = src[1];
            B = src[2];

            RGB82YUV8 (R, G, B, Y, U, V);

            ydst[0] = Y;
            udst[0] = U;
            vdst[0] = V;

            src += 3;
            ydst += 2;
            if (x % 2)
              {
                udst += 4;
                vdst += 4;
              }
          }
      }

      for (y = 0; y < CAPTURE_HEIGHT; y++)
      {
        int       x;
        const unsigned char *src = pixels + row_stride * y;

        unsigned char *dst = (unsigned char *) (fb + 0 + (y + CAPTURE_HEIGHT * y_tile) * BUFFER_WIDTH * 4) + CAPTURE_WIDTH * 4 * x_tile;

        for (x = 0; x < CAPTURE_WIDTH; x++)
          {
            dst[0] = src[0];
            dst[1] = src[1];
            dst[2] = src[2];
	    dst[3] = 255;

            src += 3;
            dst += 4;
          }
      }
}

